#include <stdexcept>
#include "gldbg.h"
#include "glprint.h"
#include "tracer.h"

using namespace std;

Tracer::Tracer(GlDbg &dbg):
	glprint(new GlPrint(0, 16384)),
	out(0),
	close_out(true),
	enabled(true)
{
	dbg.get_command_interpreter().register_command("trace", this, &Tracer::cmd_trace)
		.set_help("Traces GL function calls",
			"trace\n"
			"  Send trace output to stdout.\n\n"
			"trace FILE\n"
			"  Send trace output to FILE (- for stdout).\n\n"
			"trace {off|on}\n"
			"  Temporarily suspend or resume trace without closing the file.\n\n"
			"trace end\n"
			"  Terminate trace, closing the file.\n");
}

Tracer::~Tracer()
{
	delete glprint;
	if(close_out && out)
		fclose(out);
}

void Tracer::set_output(FILE *o, bool d)
{
	if(close_out && out)
		fclose(out);
	out = o;
	close_out = d;
}

void Tracer::decode(const char *data, unsigned len)
{
	if(!enabled || !out)
		return;

	int ret = glprint->decode(data, len);
	if(ret>=0)
	{
		const char *buf = glprint->get_buffer();
		if(buf[0])
			fprintf(out, "%s\n", buf);
	}
}

void Tracer::cmd_trace(const string &args)
{
	if(args.empty() || args=="-")
	{
		set_output(stdout, false);
		printf("Tracing to stdout\n");
	}
	else if(args=="on")
	{
		if(!out)
			throw runtime_error("Output is not set");
		enabled = true;
		printf("Tracing enabled\n");
	}
	else if(args=="off")
	{
		enabled = false;
		printf("Tracing disabled\n");
	}
	else if(args=="end")
	{
		set_output(0, true);
		printf("Tracing terminated\n");
	}
	else
	{
		FILE *f = fopen(args.c_str(), "w");
		if(!f)
			throw runtime_error("Could not open trace file");

		set_output(f, true);
		printf("Tracing to %s\n", args.c_str());
	}
}
